@@ -12,7 +12,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tkMacOSXXStubs.c,v 1.2.2.2 2004/02/23 10:49:29 das Exp $
+ * RCS: @(#) $Id: tkMacOSXXStubs.c,v 1.2.2.6 2005/05/15 20:57:08 wolfsuit Exp $
  */
 
 #include "tkInt.h"
@@ -181,13 +181,6 @@ TkpCloseDisplay(
         panic("TkpCloseDisplay: tried to call TkpCloseDisplay on bad display");
     }
 
-    /*
-     * Make sure that the local scrap is transfered to the global
-     * scrap if needed.
-     */
-
-    TkSuspendClipboard();
-
     gMacDisplay = NULL;
     if (display->screens != (Screen *) NULL) {
         if (display->screens->root_visual != (Visual *) NULL) {
@@ -198,6 +191,49 @@ TkpCloseDisplay(
     ckfree((char *) display);
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * TkClipCleanup --
+ *
+ *	This procedure is called to cleanup resources associated with
+ *	claiming clipboard ownership and for receiving selection get
+ *	results.  This function is called in tkWindow.c.  This has to be
+ *	called by the display cleanup function because we still need the
+ *	access display elements.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	Resources are freed - the clipboard may no longer be used.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+TkClipCleanup(dispPtr)
+    TkDisplay *dispPtr;	/* display associated with clipboard */
+{
+    /*
+     * Make sure that the local scrap is transfered to the global
+     * scrap if needed.
+     */
+
+    TkSuspendClipboard();
+
+    if (dispPtr->clipWindow != NULL) {
+	Tk_DeleteSelHandler(dispPtr->clipWindow, dispPtr->clipboardAtom,
+		dispPtr->applicationAtom);
+	Tk_DeleteSelHandler(dispPtr->clipWindow, dispPtr->clipboardAtom,
+		dispPtr->windowAtom);
+
+	Tk_DestroyWindow(dispPtr->clipWindow);
+	Tcl_Release((ClientData) dispPtr->clipWindow);
+	dispPtr->clipWindow = NULL;
+    }
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -329,7 +365,7 @@ XGetImage(display, d, x, y, width, height, plane_mask, format)
     int format;
 { 
     XImage *   imagePtr = NULL;
-    Pixmap     pixmap = NULL;
+    Pixmap     pixmap = (Pixmap) NULL;
     Tk_Window  win = (Tk_Window) ((MacDrawable *) d)->winPtr;
     GC         gc;
     int        depth = 32;
@@ -383,9 +419,29 @@ XGetGeometry(display, d, root_return, x_return, y_return, width_return,
     unsigned int* border_width_return;
     unsigned int* depth_return;
 {
-    /* Used in tkCanvPs.c & wm code */
-    Debugger();
-    return 0;
+    TkWindow *winPtr = ((MacDrawable *) d)->winPtr;
+
+    display->request++;
+    *root_return = ROOT_ID;
+    if (winPtr) {
+	*x_return = Tk_X(winPtr);
+	*y_return = Tk_Y(winPtr);
+	*width_return = Tk_Width(winPtr);
+	*height_return = Tk_Height(winPtr);
+	*border_width_return = winPtr->changes.border_width;
+        *depth_return = Tk_Depth(winPtr);
+    } else {
+	Rect boundsRect;
+	CGrafPtr destPort = TkMacOSXGetDrawablePort(d);
+	GetPortBounds(destPort,&boundsRect);
+	*x_return = boundsRect.left;
+	*y_return =  boundsRect.top;
+	*width_return = boundsRect.right - boundsRect.left;
+	*height_return = boundsRect.bottom - boundsRect.top;
+    	*border_width_return = 0;
+        *depth_return = 32;
+    }
+    return 1;
 }
 
 void
@@ -705,10 +761,11 @@ TkMacOSXXGetPixel(
     int x,
     int y)
 {
-    CGrafPtr grafPtr;
+    CGrafPtr grafPtr, oldPort;
     RGBColor cPix;
     unsigned long r, g, b, c;
     grafPtr = (CGrafPtr)image->data;
+    GetPort(&oldPort);
     SetPort(grafPtr);
     GetCPixel(x,y,&cPix);
     if (image->obdata) {
@@ -722,6 +779,7 @@ TkMacOSXXGetPixel(
         b = cPix . blue;
     }
     c = (r<<16)|(g<<8)|(b);
+    SetPort(oldPort);
     return c;
 }
 
@@ -732,10 +790,11 @@ TkMacOSXXPutPixel(
     int y,
     unsigned long pixel)
 {
-    CGrafPtr grafPtr;
+    CGrafPtr grafPtr, oldPort;
     RGBColor cPix;
     unsigned long r, g, b;
     grafPtr = (CGrafPtr)image->data;
+    GetPort(&oldPort);
     SetPort(grafPtr);
     r  = (pixel & image->red_mask)>>16;
     g  = (pixel & image->green_mask)>>8;
@@ -751,6 +810,7 @@ TkMacOSXXPutPixel(
         cPix . blue = b;
     }
     SetCPixel(x,y,&cPix);
+    SetPort(oldPort);
     return 0;
 }
 
